// point2d.h -- a 2d point class.  
//
// Adapted from point3d.hxx by Curtis L. Olson
// started June 1999.
//
// Copyright (C) 1999  Christian E. Mayer  - vader@t-online.de
//
// This program is free software; you can redistribute it and/or
// modify it under the terms of the GNU General Public License as
// published by the Free Software Foundation; either version 2 of the
// License, or (at your option) any later version.
//
// This program is distributed in the hope that it will be useful, but
// WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
// General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with this program; if not, write to the Free Software
// Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.


#ifndef _POINT2D_H
#define _POINT2D_H


#ifndef __cplusplus                                                          
# error This library requires C++
#endif                                   

#include "compiler.h"
#include <Math/point3d.hxx>

#ifdef FG_MATH_EXCEPTION_CLASH
# define exception c_exception
#endif

#ifdef FG_HAVE_STD_INCLUDES
# include <iostream>
# include <cassert>
# include <cmath>
#else
# include <iostream.h>
# include <assert.h>
# include <math.h>
#endif

#ifndef FG_HAVE_NATIVE_SGI_COMPILERS
FG_USING_STD(ostream);
FG_USING_STD(istream);
#endif

// -dw- someone seems to have forgotten this...
#ifdef __MWERKS__
FG_USING_STD(std);
#endif


const double fgPoint2_Epsilon = 0.0000001;

//the 2d is needed to avoid problems with point3d.hxx
enum {PX2d, PY2d};			// axes

// Kludge for msvc++ 6.0 - requires forward decls of friend functions.
class Point2D;
istream& operator>> ( istream&, Point2D& );
ostream& operator<< ( ostream&, const Point2D& );
Point2D operator- (const Point2D& p);	            // -p1
bool operator== (const Point2D& a, const Point2D& b);  // p1 == p2?


///////////////////////////
//
// 2D Point
//
///////////////////////////

class Point2D {

protected:

    double n[2];

public:

    // Constructors

    Point2D();
    Point2D(const double x, const double y);
    explicit Point2D(const double d);
    Point2D(const Point2D &p);
    Point2D(const Point3D &p);

    // Assignment operators

    Point2D& operator = ( const Point2D& p );	 // assignment of a Point3D
    Point2D& operator += ( const Point2D& p );	 // incrementation by a Point3D
    Point2D& operator -= ( const Point2D& p );	 // decrementation by a Point3D
    Point2D& operator *= ( const double d );     // multiplication by a constant
    Point2D& operator /= ( const double d );	 // division by a constant

    void setx(const double x);
    void sety(const double y);

    // Queries 

    double& operator [] ( int i);		 // indexing
    double operator[] (int i) const;		 // read-only indexing

    double x() const;      // cartesian x
    double y() const;      // cartesian y

    // friends
    friend Point2D operator - (const Point2D& p);	            // -p1
    friend bool operator == (const Point2D& a, const Point2D& b);  // p1 == p2?
    friend istream& operator>> ( istream&, Point2D& );
    friend ostream& operator<< ( ostream&, const Point2D& );

    // Special functions
    double distance2D(const Point2D& a) const;        // distance between
    double distance2Dsquared(const Point2D& a) const; // distance between ^ 2
    double distance3D(const Point2D& a) const;        // distance between
    double distance3Dsquared(const Point2D& a) const; // distance between ^ 2
};


// input from stream
inline istream&
operator >> ( istream& in, Point2D& p)
{
    char c;

    in >> p.n[PX2d];

    // read past optional comma
    while ( in.get(c) ) {
	if ( (c != ' ') && (c != ',') ) {
	    // push back on the stream
	    in.putback(c);
	    break;
	}
    }
	
    in >> p.n[PY2d];

    return in;
}

inline ostream&
operator<< ( ostream& out, const Point2D& p )
{
    return out << p.n[PX2d] << ", " << p.n[PY2d];
}

///////////////////////////
//
// Point2D Member functions
//
///////////////////////////

// CONSTRUCTORS

inline Point2D::Point2D() {}

inline Point2D::Point2D(const double x, const double y)
{
    n[PX2d] = x; n[PY2d] = y;
}

inline Point2D::Point2D(const double d)
{
    n[PX2d] = n[PY2d] = d;
}

inline Point2D::Point2D(const Point3D &p)
{
    n[PX2d] = p.x(); n[PY2d] = p.y();
}

inline Point2D::Point2D(const Point2D& p)
{
    n[PX2d] = p.n[PX2d]; n[PY2d] = p.n[PY2d];
}

// ASSIGNMENT OPERATORS

inline Point2D& Point2D::operator = (const Point2D& p)
{
    n[PX2d] = p.n[PX2d]; n[PY2d] = p.n[PY2d]; return *this;
}

inline Point2D& Point2D::operator += ( const Point2D& p )
{
    n[PX2d] += p.n[PX2d]; n[PY2d] += p.n[PY2d]; return *this;
}

inline Point2D& Point2D::operator -= ( const Point2D& p )
{
    n[PX2d] -= p.n[PX2d]; n[PY2d] -= p.n[PY2d]; return *this;
}

inline Point2D& Point2D::operator *= ( const double d )
{
    n[PX2d] *= d; n[PY2d] *= d; return *this;
}

inline Point2D& Point2D::operator /= ( const double d )
{
    double d_inv = 1./d; n[PX2d] *= d_inv; n[PY2d] *= d_inv;
    return *this;
}

inline void Point2D::setx(const double x) {
    n[PX2d] = x;
}

inline void Point2D::sety(const double y) {
    n[PY2d] = y;
}

// QUERIES

inline double& Point2D::operator [] ( int i)
{
    assert(! (i < PX2d || i > PY2d));
    return n[i];
}

inline double Point2D::operator [] ( int i) const {
    assert(! (i < PX2d || i > PY2d));
    return n[i];
}


inline double Point2D::x() const { return n[PX2d]; }

inline double Point2D::y() const { return n[PY2d]; }


// FRIENDS

inline Point2D operator - (const Point2D& a)
{
    return Point2D(-a.n[PX2d],-a.n[PY2d]);
}

inline Point2D operator + (const Point2D& a, const Point2D& b)
{
    return Point2D(a) += b;
}

inline Point2D operator - (const Point2D& a, const Point2D& b)
{
    return Point2D(a) -= b;
}

inline Point2D operator * (const Point2D& a, const double d)
{
    return Point2D(a) *= d;
}

inline Point2D operator * (const double d, const Point2D& a)
{
    return a*d;
}

inline Point2D operator / (const Point2D& a, const double d)
{
    return Point2D(a) *= (1.0 / d );
}

#ifndef ABS
#   define ABS(x) (((x)>0)?(x):(-(x)))
#endif

inline bool operator == (const Point2D& a, const Point2D& b)
{
    return
	(( ABS(a.n[PX2d] - b.n[PX2d]) < fgPoint2_Epsilon ) &&
	 ( ABS(a.n[PY2d] - b.n[PY2d]) < fgPoint2_Epsilon )    );
}

inline bool operator != (const Point2D& a, const Point2D& b)
{
    return !(a == b);
}

// Special functions

inline double
Point2D::distance2D(const Point2D& a ) const
{
    double x, y;

    x = n[PX2d] - a.n[PX2d];
    y = n[PY2d] - a.n[PY2d];

    return sqrt(x*x + y*y);
}


inline double
Point2D::distance2Dsquared(const Point2D& a ) const
{
    double x, y;

    x = n[PX2d] - a.n[PX2d];
    y = n[PY2d] - a.n[PY2d];

    return(x*x + y*y);
}

inline double
Point2D::distance3D(const Point2D& a ) const
{
    double x, y;

    x = n[PX2d] - a.n[PX2d];
    y = n[PY2d] - a.n[PY2d];

    return sqrt(x*x + y*y);
}


inline double
Point2D::distance3Dsquared(const Point2D& a ) const
{
    double x, y;

    x = n[PX2d] - a.n[PX2d];
    y = n[PY2d] - a.n[PY2d];

    return(x*x + y*y);
}



#endif // _POINT2D_HXX


